defmodule Hafnium.Utils do
  @moduledoc false
  # @dialyzer {:nowarn_function, stringify_keys: 1, atomize_keys: 1}

  @doc """
  Stringify map keys.

  ## Examples

    iex> #{__MODULE__}.stringify_keys(%{})
    %{}
    iex> #{__MODULE__}.stringify_keys(%{foo: "bar"})
    %{"foo" => "bar"}
    iex> #{__MODULE__}.stringify_keys(%{foo: :bar})
    %{"foo" => :bar}
    iex> #{__MODULE__}.stringify_keys(%{"foo" => "bar"})
    %{"foo" => "bar"}

  """
  @spec stringify_keys(map()) :: map()
  def stringify_keys(%{} = map) do
    map
    |> Enum.map(fn {k, v} -> {inner_stringify_keys(k), v} end)
    |> Enum.into(%{})
  end

  defp inner_stringify_keys(key) when is_atom(key), do: Atom.to_string(key)
  defp inner_stringify_keys(key), do: key

  @doc """
  Atomize map keys.

  ## Examples

      iex> #{__MODULE__}.atomize_keys(%{})
      %{}
      iex> #{__MODULE__}.atomize_keys(%{foo: "bar"})
      %{foo: "bar"}
      iex> #{__MODULE__}.atomize_keys(%{foo: :bar})
      %{foo: :bar}
      iex> #{__MODULE__}.atomize_keys(%{"foo" => "bar"})
      %{foo: "bar"}
  """
  @spec atomize_keys(map()) :: map()
  def atomize_keys(%{} = map) do
    map
    |> Enum.map(fn {k, v} -> {inner_atomize_keys(k), v} end)
    |> Enum.into(%{})
  end

  defp inner_atomize_keys(key) when is_binary(key), do: String.to_atom(key)
  defp inner_atomize_keys(key), do: key

  @doc """
  Get module name

  ## Examples

      iex> #{__MODULE__}.module_name(Foo.Bar.FooBar)
      :foo_bar_foo_bar
  """
  @spec module_name(module()) :: atom()
  def module_name(module) do
    module
    |> to_string
    |> String.trim_leading("Elixir.")
    |> String.replace(".", "")
    |> Macro.underscore()
    |> String.to_atom()
  end

  @spec all_modules :: [atom()]
  def all_modules do
    Enum.reduce(Application.started_applications(), [], fn {app, _descr, _vsn}, acc ->
      case :application.get_key(app, :modules) do
        {:ok, modules} ->
          modules ++ acc

        _ ->
          acc
      end
    end)
  end

  @spec defined?(any()) :: boolean()
  def defined?(module) when is_atom(module) do
    module
    |> Code.ensure_compiled()
    |> case do
      {:module, _} -> true
      _ -> false
    end
  end

  def defined?(_), do: false
end
